/*
 * Copyright 2016 - 2023 Gary Talent (gary@drinkingtea.net). All rights reserved.
 */

#include <GLFW/glfw3.h>
#include <imgui_impl_opengl3.h>
#include <imgui_impl_glfw.h>

#include <keel/keel.hpp>
#include <turbine/gfx.hpp>

#include "../config.hpp"
#include "context.hpp"

namespace turbine {

ox::Result<ox::UPtr<Context>> init(ox::UPtr<ox::FileSystem> fs, ox::CRStringView appName) noexcept {
	auto ctx = ox::make_unique<GlfwContext>();
	oxReturnError(keel::init(&ctx->keelCtx, std::move(fs), appName));
	using namespace std::chrono;
	ctx->startTime = duration_cast<milliseconds>(system_clock::now().time_since_epoch()).count();
	glfwInit();
	oxReturnError(initGfx(*ctx));
	return ox::UPtr<Context>(ctx.release());
}

static void tickFps(GlfwContext &gctx, uint64_t nowMs) noexcept {
	++gctx.draws;
	if (gctx.draws >= 500) {
		const auto duration = static_cast<double>(nowMs - gctx.prevFpsCheckTime) / 1000.0;
		const auto fps = static_cast<int>(static_cast<double>(gctx.draws) / duration);
		if constexpr(config::GlFpsPrint) {
			oxOutf("FPS: {}\n", fps);
		}
		oxTracef("turbine::fps", "FPS: {}", fps);
		gctx.prevFpsCheckTime = nowMs;
		gctx.draws = 0;
	}
}

ox::Error run(Context &ctx) noexcept {
	auto &gctx = static_cast<GlfwContext&>(ctx);
	int sleepTime = 0;
	while (!glfwWindowShouldClose(gctx.window)) {
		glfwPollEvents();
		const auto ticks = ticksMs(gctx);
		if (gctx.wakeupTime <= ticks) {
			 sleepTime = gctx.updateHandler(gctx);
			 if (sleepTime >= 0) {
				  gctx.wakeupTime = ticks + static_cast<unsigned>(sleepTime);
			 } else {
				  gctx.wakeupTime = ~uint64_t(0);
			 }
		} else {
			sleepTime = 10;
		}
		tickFps(gctx, ticks);
		// draw start
		if constexpr(turbine::config::ImGuiEnabled) {
			ImGui_ImplOpenGL3_NewFrame();
			ImGui_ImplGlfw_NewFrame();
			ImGui::NewFrame();
		}
		for (auto d : gctx.drawers) {
			d->draw(gctx);
		}
		if constexpr(turbine::config::ImGuiEnabled) {
			ImGui::Render();
			ImGui_ImplOpenGL3_RenderDrawData(ImGui::GetDrawData());
		}
		// draw end
		glfwSwapBuffers(gctx.window);
		if (!gctx.constantRefresh) {
			if (gctx.uninterruptedRefreshes) {
				--gctx.uninterruptedRefreshes;
			} else {
				glfwWaitEventsTimeout(sleepTime);
			}
		}
	}
	shutdown(gctx);
	return {};
}

void shutdown(Context &ctx) noexcept {
	auto &gctx = static_cast<GlfwContext&>(ctx);
	if (gctx.window) {
		ImGui_ImplOpenGL3_Shutdown();
		ImGui_ImplGlfw_Shutdown();
		glfwDestroyWindow(gctx.window);
		gctx.window = nullptr;
	}
}

uint64_t ticksMs(const Context &ctx) noexcept {
	using namespace std::chrono;
	auto &gctx = static_cast<const GlfwContext&>(ctx);
	const auto now = duration_cast<milliseconds>(system_clock::now().time_since_epoch()).count();
	return static_cast<uint64_t>(now - gctx.startTime);
}

bool buttonDown(const Context &ctx, Key key) noexcept {
	auto &gctx = static_cast<const GlfwContext&>(ctx);
	return (gctx.keysDown >> static_cast<int>(key)) & 1;
}

void requestShutdown(Context &ctx) noexcept {
	auto &gctx = static_cast<GlfwContext&>(ctx);
	glfwSetWindowShouldClose(gctx.window, true);
}

}
